/*=============================================================================
Vitalnix User Management Suite
backends/multi.c
  Copyright © Jan Engelhardt <jengelh [at] gmx de>, 2006
  -- License restrictions apply (LGPL v2.1)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA

  -- For details, see the file named "LICENSE.LGPL2"
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <libHX.h>
#include "vitalnix-config.h"
#include "backends/static-build.h"
#include "libvxpdb/libvxpdb.h"
#include "libvxpdb/xafunc.h"

#define WR_OPEN(state) ((state)->wr_mod.mh_state == STATE_OPEN)
#define WR_MOD(state) ((state)->wr_mod.mh_instance)

// Definitions
enum open_state {
    STATE_OUT,
    STATE_LOADED,
    STATE_OPEN,
};

struct module_handle {
    char *mh_name;
    enum open_state mh_state;
    struct vxpdb_state *mh_instance;
};

struct multi_state {
    int destructing;
    struct HXdeque *rd_mod;
    struct module_handle wr_mod;
};

// Functions
static int vmulti_init(struct vxpdb_state *, const void *);
static int vmulti_open(struct vxpdb_state *, long);
static void vmulti_close(struct vxpdb_state *);
static long vmulti_modctl(struct vxpdb_state *, long, ...);
static void vmulti_deinit(struct vxpdb_state *);
static int vmulti_lock(struct vxpdb_state *);
static int vmulti_unlock(struct vxpdb_state *);

static int vmulti_useradd(struct vxpdb_state *, const struct vxpdb_user *);
static int vmulti_usermod(struct vxpdb_state *, const struct vxpdb_user *, const struct vxpdb_user *);
static int vmulti_userdel(struct vxpdb_state *, const struct vxpdb_user *);
static int vmulti_userinfo(struct vxpdb_state *, const struct vxpdb_user *, struct vxpdb_user *, size_t);
static void *vmulti_usertrav_init(struct vxpdb_state *);
static int vmulti_usertrav_walk(struct vxpdb_state *, void *, struct vxpdb_user *);
static void vmulti_usertrav_free(struct vxpdb_state *, void *);

static int vmulti_groupadd(struct vxpdb_state *, const struct vxpdb_group *);
static int vmulti_groupmod(struct vxpdb_state *, const struct vxpdb_group *, const struct vxpdb_group *);
static int vmulti_groupdel(struct vxpdb_state *, const struct vxpdb_group *);
static int vmulti_groupinfo(struct vxpdb_state *, const struct vxpdb_group *, struct vxpdb_group *, size_t);
static void *vmulti_grouptrav_init(struct vxpdb_state *);
static int vmulti_grouptrav_walk(struct vxpdb_state *, void *, struct vxpdb_group *);
static void vmulti_grouptrav_free(struct vxpdb_state *, void *);

static int modules_construct(struct multi_state *, char **);
static int modules_open(struct multi_state *, long);
static void modules_close(struct multi_state *);
static void modules_destruct(struct multi_state *);
static void read_config(struct multi_state *);
static void read_rdmod(const struct HXoptcb *);
static void read_wrmod(const struct HXoptcb *);

// Variables
COND_STATIC struct vxpdb_module THIS_MODULE = {
    .name           = "Multiple Module driver",
    .author         = "Jan Engelhardt <jengelh [at] gmx de>, 2006",

    .init           = vmulti_init,
    .open           = vmulti_open,
    .close          = vmulti_close,
    .deinit         = vmulti_deinit,
    .modctl         = vmulti_modctl,
    .lock           = vmulti_lock,
    .unlock         = vmulti_unlock,

    .useradd        = vmulti_useradd,
    .usermod        = vmulti_usermod,
    .userdel        = vmulti_userdel,
    .userinfo       = vmulti_userinfo,
    .usertrav_init  = vmulti_usertrav_init,
    .usertrav_walk  = vmulti_usertrav_walk,
    .usertrav_free  = vmulti_usertrav_free,

    .groupadd       = vmulti_groupadd,
    .groupmod       = vmulti_groupmod,
    .groupdel       = vmulti_groupdel,
    .groupinfo      = vmulti_groupinfo,
    .grouptrav_init = vmulti_grouptrav_init,
    .grouptrav_walk = vmulti_grouptrav_walk,
    .grouptrav_free = vmulti_grouptrav_free,
};

//-----------------------------------------------------------------------------
REGISTER_MODULE(multi, &THIS_MODULE);

static int vmulti_init(struct vxpdb_state *this, const void *private_data) {
    struct multi_state *state;

    if((state = this->state = calloc(1, sizeof(struct multi_state))) == NULL)
        return -errno;

    read_config(state);
    return 1;
}

static int vmulti_open(struct vxpdb_state *this, long flags) {
    struct multi_state *state = this->state;
    int ret;

    if((ret = modules_open(state, flags)) <= 0)
        return ret;

    return 1;
}

static void vmulti_close(struct vxpdb_state *this) {
    struct multi_state *state = this->state;
    modules_close(state);
    free(state);
    return;
}

static void vmulti_deinit(struct vxpdb_state *this) {
    struct multi_state *state = this->state;
    modules_destruct(state);
    free(state);
    return;
}

static long vmulti_modctl(struct vxpdb_state *this, long command, ...) {
    return -ENOSYS;
}

static int vmulti_lock(struct vxpdb_state *this) {
    struct multi_state *state = this->state;
    if(WR_OPEN(state))
        return vxpdb_lock(WR_MOD(state));
    return 1;
}

static int vmulti_unlock(struct vxpdb_state *this) {
    struct multi_state *state = this->state;
    if(WR_OPEN(state))
        return vxpdb_unlock(WR_MOD(state));
    return 1;
}

//-----------------------------------------------------------------------------
static int vmulti_useradd(struct vxpdb_state *this, const struct vxpdb_user *u) {
    struct multi_state *state = this->state;
    if(WR_OPEN(state))
        return vxpdb_useradd(WR_MOD(state), u);
    return -EROFS;
}

static int vmulti_usermod(struct vxpdb_state *this,
 const struct vxpdb_user *search, const struct vxpdb_user *modify)
{
    struct multi_state *state = this->state;
    if(WR_OPEN(state))
        return vxpdb_usermod(WR_MOD(state), search, modify);
    return -EROFS;
}

static int vmulti_userdel(struct vxpdb_state *this, const struct vxpdb_user *sa) {
    struct multi_state *state = this->state;
    if(WR_OPEN(state))
        return vxpdb_userdel(WR_MOD(state), sa);
    return -EROFS;
}

static int vmulti_userinfo(struct vxpdb_state *this,
 const struct vxpdb_user *mask, struct vxpdb_user *result, size_t size)
{
/*    struct shadow_state *state = this->state;
    const struct HXdeque_node *travp = state->dq_user->first;
    struct vxpdb_user temp_mask;
    int found = 0;

    if(mask == result) {
        memcpy(&temp_mask, mask, sizeof(struct vxpdb_user));
        mask = &temp_mask;
    }

    while(travp != NULL && (result == NULL || (result != NULL && size > 0))) {
        const struct vxpdb_user *src = travp->ptr;

        if(!vxpdb_user_match(src, mask)) {
            travp = travp->Next;
            continue;
        }

        if(result != NULL) {
            vxpdb_user_copy(result, src);
            ++result;
            ++found;
            --size;
        } else {
            if(size == 0)
                return 1;
            ++found;
        }
        travp = travp->Next;
    }

    return found;*/
    return 0;
}

static void *vmulti_usertrav_init(struct vxpdb_state *this) {
    return NULL;
}

static int vmulti_usertrav_walk(struct vxpdb_state *this, void *ptr,
 struct vxpdb_user *result)
{
    return 1;
}

static void vmulti_usertrav_free(struct vxpdb_state *this, void *ptr) {
    return;
}

//-----------------------------------------------------------------------------
static int vmulti_groupadd(struct vxpdb_state *this, const struct vxpdb_group *g) {
    struct multi_state *state = this->state;
    if(WR_OPEN(state))
        return vxpdb_groupadd(WR_MOD(state), g);
    return -EROFS;
}

static int vmulti_groupmod(struct vxpdb_state *this,
 const struct vxpdb_group *sr_mask, const struct vxpdb_group *mod_mask)
{
    struct multi_state *state = this->state;
    if(WR_OPEN(state))
        return vxpdb_groupmod(WR_MOD(state), sr_mask, mod_mask);
    return -EROFS;
}

static int vmulti_groupdel(struct vxpdb_state *this,
 const struct vxpdb_group *sr_mask)
{
    struct multi_state *state = this->state;
    if(WR_OPEN(state))
        return vxpdb_groupdel(WR_MOD(state), sr_mask);
    return -EROFS;
}

static void *vmulti_grouptrav_init(struct vxpdb_state *this) {
    return NULL;
}

static int vmulti_grouptrav_walk(struct vxpdb_state *this, void *ptr,
 struct vxpdb_group *result)
{
    return 1;
}

static void vmulti_grouptrav_free(struct vxpdb_state *this, void *ptr) {
    return;
}

static int vmulti_groupinfo(struct vxpdb_state *this, const struct vxpdb_group *rq,
 struct vxpdb_group *dest, size_t size)
{
    return 0;
}

//-----------------------------------------------------------------------------
static int modules_construct(struct multi_state *state, char **name) {
    state->rd_mod = HXdeque_init();

    while(name != NULL && *name != NULL) {
        struct module_handle mh = {
            .mh_name     = *name,
            .mh_state    = STATE_OUT,
        };
        HXdeque_push(state->rd_mod, HX_memdup(&mh, sizeof(mh)));
    }

    return 1;
}

static int modules_open(struct multi_state *state, long flags) {
    struct HXdeque_node *nd = state->rd_mod->first;
    int ret = 1;

    while(nd != NULL) {
        struct module_handle *mh = nd->ptr;
        if((ret = vxpdb_open(mh->mh_instance, flags & ~PDB_WRLOCK)) <= 0) {
            modules_close(state);
            return ret;
        } else {
            mh->mh_state = STATE_OPEN;
        }
        nd = nd->Next;
    }

    return ret;
}

static void modules_close(struct multi_state *state) {
    struct HXdeque_node *nd = state->rd_mod->first;

    while(nd != NULL) {
        struct module_handle *mod = nd->ptr;
        if(mod->mh_state == STATE_OPEN)
            vxpdb_close(mod->mh_instance);
        mod->mh_state = STATE_LOADED;
    }

    if(WR_OPEN(state)) {
        vxpdb_close(WR_MOD(state));
        state->wr_mod.mh_state = STATE_LOADED;
    }

    return;
}

static void modules_destruct(struct multi_state *state) {
    struct HXdeque_node *nd = state->rd_mod->first;

    while(nd != NULL) {
        struct module_handle *mod = nd->ptr;
        if(mod->mh_state == STATE_LOADED) {
            vxpdb_unload(mod->mh_instance);
            mod->mh_state = STATE_OUT;
        }
        nd = nd->Next;
    }

    if(state->wr_mod.mh_state == STATE_LOADED) {
        vxpdb_unload(WR_MOD(state));
        state->wr_mod.mh_state = STATE_OUT;
    }

    return;
}

static void read_config(struct multi_state *state) {
    struct HXoption options_table[] = {
        {.ln = "RD_MODULES", .type = HXTYPE_STRING, .cb = read_rdmod, .uptr = state},
        {.ln = "WR_MODULE",  .type = HXTYPE_STRING, .cb = read_wrmod, .uptr = state},
        HXOPT_TABLEEND,
    };

    HX_shconfig(CONFIG_ETC_VITALNIX "/db_multi.conf", options_table);
    return;
}

static void read_rdmod(const struct HXoptcb *cbi) {
    struct multi_state *state = cbi->current->uptr;
    char **mod = HX_split(cbi->s, ":", NULL, 0);
    if((state->destructing = modules_construct(state, mod)) <= 0)
        modules_destruct(state);
    HX_zvecfree(mod);
    return;
}

static void read_wrmod(const struct HXoptcb *cbi) {
    struct multi_state *state = cbi->current->uptr;
    if(state->destructing == 0) {
        state->wr_mod.mh_name  = HX_strdup(cbi->s);
        state->wr_mod.mh_state = STATE_OUT;
    }
    return;
}

//=============================================================================
