/*=============================================================================
Vitalnix User Management Suite
backends/shadow.c - Shadow back-end module
  Copyright © Jan Engelhardt <jengelh [at] gmx de>, 2002 - 2006
  -- License restrictions apply (LGPL v2.1)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA

  -- For details, see the file named "LICENSE.LGPL2"
=============================================================================*/
#include <sys/file.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libHX.h>
#include "vitalnix-config.h"
#include "backends/static-build.h"
#include "libvxpdb/libvxpdb.h"
#include "libvxutil/defines.h"
#include "libvxutil/libvxutil.h"

#define RWLOCKED(state)         ((state)->flags & PDB_WRLOCK)
#define RWLOCK_CHK(state)       if(!RWLOCKED(state)) return -EPERM;
#define TOUCH_GROUP_TAG(i)      (state->fgroup.is_chg = (i))
#define TOUCH_USER_TAG(i)       \
    (state->fpasswd.is_chg = state->fshadow.is_chg = \
    state->fvxshadow.is_chg = (i))

// Definitions
enum {
    IDF_NORMAL = 0,
    IDF_SYSTEM,
    CONFIG_READ  = 0,
    CONFIG_FREE,
    FLUSH_AUTO = 0,
    FLUSH_FORCE,
};

struct spdb_file {
    char *path;
    FILE *fp;
    int fd;
    char is_chg;
};

struct shadow_state {
    unsigned long flags;
    struct HXdeque *dq_user, *dq_group;
    struct spdb_file fpasswd, fshadow, fvxpasswd, fvxshadow, fgroup;
    long uid_min, uid_max, gid_min, gid_max;
};

struct traverser_state {
    struct HXdeque_node *wp;
};

// Functions
static int vshadow_init(struct vxpdb_state *, const void *);
static int vshadow_open(struct vxpdb_state *, long);
static void vshadow_close(struct vxpdb_state *);
static void vshadow_deinit(struct vxpdb_state *);
static long vshadow_modctl(struct vxpdb_state *, long, ...);
static int vshadow_lock(struct vxpdb_state *);
static int vshadow_unlock(struct vxpdb_state *);

static int vshadow_useradd(struct vxpdb_state *, const struct vxpdb_user *);
static int vshadow_usermod(struct vxpdb_state *, const struct vxpdb_user *, const struct vxpdb_user *);
static int vshadow_userdel(struct vxpdb_state *, const struct vxpdb_user *);
static int vshadow_userinfo(struct vxpdb_state *, const struct vxpdb_user *, struct vxpdb_user *, size_t);
static void *vshadow_usertrav_init(struct vxpdb_state *);
static int vshadow_usertrav_walk(struct vxpdb_state *, void *, struct vxpdb_user *);
static void vshadow_usertrav_free(struct vxpdb_state *, void *);

static int vshadow_groupadd(struct vxpdb_state *, const struct vxpdb_group *);
static int vshadow_groupmod(struct vxpdb_state *, const struct vxpdb_group *, const struct vxpdb_group *);
static int vshadow_groupdel(struct vxpdb_state *, const struct vxpdb_group *);
static int vshadow_groupinfo(struct vxpdb_state *, const struct vxpdb_group *, struct vxpdb_group *, size_t);
static void *vshadow_grouptrav_init(struct vxpdb_state *);
static int vshadow_grouptrav_walk(struct vxpdb_state *, void *, struct vxpdb_group *);
static void vshadow_grouptrav_free(struct vxpdb_state *, void *);

static int db_open(struct shadow_state *, int);
static void db_close(struct shadow_state *);
static struct HXdeque *db_read_passwd(FILE *);
static void db_read_shadow(FILE *, struct HXdeque *);
static void db_read_vxshadow(FILE *, struct HXdeque *);
static struct HXdeque *db_read_groups(FILE *);
static void db_flush(struct shadow_state *, int);
static void db_flush_users(struct shadow_state *);
static void db_flush_groups(struct shadow_state *);
static void db_write_passwd(FILE *, const struct vxpdb_user *);
static void db_write_shadow(FILE *, const struct vxpdb_user *);
static inline void db_write_vxshadow(FILE *, const struct vxpdb_user *);

static long automatic_uid(struct shadow_state *, long);
static long automatic_gid(struct shadow_state *, long);
static void free_data(struct shadow_state *);
static inline void free_single_group(struct vxpdb_group *);
static inline void free_single_user(struct vxpdb_user *);
static inline struct vxpdb_group *lookup_group(struct HXdeque *, const char *, long);
static inline struct vxpdb_user *lookup_user(struct HXdeque *, const char *, long);
static void read_config(struct shadow_state *, unsigned int);
static struct HXdeque_node *skip_nis_users(struct HXdeque_node *);
static struct HXdeque_node *skip_nis_groups(struct HXdeque_node *);
static void truncate_here(FILE *);

// Variables
COND_STATIC struct vxpdb_module THIS_MODULE = {
    .name           = "vxShadow back-end module",
    .desc           = "for shadow suite (and vxshadow extension)",
    .author         = "Jan Engelhardt <jengelh [at] gmx de>, 2002 - 2006",

    .init           = vshadow_init,
    .open           = vshadow_open,
    .close          = vshadow_close,
    .deinit         = vshadow_deinit,
    .modctl         = vshadow_modctl,
    .lock           = vshadow_lock,
    .unlock         = vshadow_unlock,

    .useradd        = vshadow_useradd,
    .usermod        = vshadow_usermod,
    .userdel        = vshadow_userdel,
    .userinfo       = vshadow_userinfo,
    .usertrav_init  = vshadow_usertrav_init,
    .usertrav_walk  = vshadow_usertrav_walk,
    .usertrav_free  = vshadow_usertrav_free,

    .groupadd       = vshadow_groupadd,
    .groupmod       = vshadow_groupmod,
    .groupdel       = vshadow_groupdel,
    .groupinfo      = vshadow_groupinfo,
    .grouptrav_init = vshadow_grouptrav_init,
    .grouptrav_walk = vshadow_grouptrav_walk,
    .grouptrav_free = vshadow_grouptrav_free,
};

static char *Path_passwd   = "/etc/passwd",
            *Path_shadow   = "/etc/shadow",
            *Path_group    = "/etc/group",
            *Path_vxpasswd = "/etc/vxpasswd",
            *Path_vxshadow = "/etc/vxshadow";

//-----------------------------------------------------------------------------
REGISTER_MODULE(shadow, &THIS_MODULE);

static int vshadow_init(struct vxpdb_state *this, const void *private_data) {
    struct shadow_state *state;

    if((state = this->state = calloc(1, sizeof(struct shadow_state))) == NULL)
        return -errno;

    state->fpasswd.path   = HX_strdup(Path_passwd);
    state->fshadow.path   = HX_strdup(Path_shadow);
    state->fgroup.path    = HX_strdup(Path_group);
    state->fvxpasswd.path = HX_strdup(Path_vxpasswd);
    state->fvxshadow.path = HX_strdup(Path_vxshadow);
    state->fpasswd.fd     = -1;
    state->fshadow.fd     = -1;
    state->fgroup.fd      = -1;
    state->fvxpasswd.fd   = -1;
    state->fvxshadow.fd   = -1;
    read_config(state, CONFIG_READ);

    if(geteuid() == 0 || getuid() == 0) {
        fprintf(stderr, "This is BETA software. Do not use it as root!\n");
        return -EPERM;
    }

    return 1;
}

static int vshadow_open(struct vxpdb_state *this, long flags) {
    struct shadow_state *state = this->state;
    int ret;

    if((ret = db_open(state, flags & PDB_WRLOCK)) <= 0) {
        vshadow_close(this);
        return ret;
    }
    return 1;
}

static void vshadow_close(struct vxpdb_state *this) {
    struct shadow_state *state = this->state;
    db_close(state);
    free_data(state);
    return;
}

static void vshadow_deinit(struct vxpdb_state *this) {
    struct shadow_state *state = this->state;
    read_config(state, CONFIG_FREE);
    free(state);
    return;
}

static long vshadow_modctl(struct vxpdb_state *this, long command, ...) {
    struct shadow_state *state = this->state;
    errno = 0;

    switch(command) {
        case PDB_FLUSH:
            if(!(state->flags & PDB_WRLOCK))
                return -EPERM;
            db_flush(state, FLUSH_FORCE);
            return 1;
        case PDB_COUNT_USERS:
            return state->dq_user->itemcount;
        case PDB_COUNT_GROUPS:
            return state->dq_group->itemcount;
        case PDB_NEXTUID_SYS:
            return automatic_uid(state, PDB_AUTOUID_SYS);
        case PDB_NEXTUID:
            return automatic_uid(state, PDB_AUTOUID);
        case PDB_NEXTGID_SYS:
            return automatic_gid(state, PDB_AUTOGID_SYS);
        case PDB_NEXTGID:
            return automatic_gid(state, PDB_AUTOGID);
    }

    return -ENOSYS;
}

static int vshadow_lock(struct vxpdb_state *this) {
    struct shadow_state *state = this->state;
    if(RWLOCKED(state))
        return 1;
    db_close(state);
    free_data(state);
    return db_open(state, state->flags | PDB_WRLOCK);
}

static int vshadow_unlock(struct vxpdb_state *this) {
    struct shadow_state *state = this->state;
    struct flock lockinfo = {
        .l_type   = F_UNLCK,
        .l_whence = SEEK_SET,
        .l_start  = 0,
        .l_len    = 0,
    };
    if(!RWLOCKED(state))
        return 1;
    if(fcntl(state->fpasswd.fd, F_SETLK, &lockinfo) != 0)
        return -errno;
    state->flags &= ~PDB_WRLOCK;
    return 1;
}

//-----------------------------------------------------------------------------
static int vshadow_useradd(struct vxpdb_state *this,
 const struct vxpdb_user *rq)
{
    struct shadow_state *state = this->state;
    struct vxpdb_user *nu;

    if(rq->pw_name == NULL)
        return -EINVAL;
    if(lookup_user(state->dq_user, rq->pw_name, PDB_NOUID) != NULL)
        return -EEXIST;

    RWLOCK_CHK(state);

    /* We have to be careful with strings here, as we need to duplicate them
    for us, leave the original ones intact and not have a single foreign
    pointer in our structs. */
    if((nu = calloc(1, sizeof(struct vxpdb_user))) == NULL)
        return -errno;

    nu->pw_name    = HX_strdup(rq->pw_name);
    nu->pw_uid     = automatic_uid(state, rq->pw_uid);
    nu->pw_gid     = rq->pw_gid;
    nu->pw_real    = HX_strdup(rq->pw_real);
    nu->pw_home    = HX_strdup(rq->pw_home);
    nu->pw_shell   = HX_strdup(rq->pw_shell);

    nu->sp_passwd  = HX_strdup(rq->sp_passwd);
    nu->sp_lastchg = rq->sp_lastchg;
    nu->sp_min     = rq->sp_min;
    nu->sp_max     = rq->sp_max;
    nu->sp_warn    = rq->sp_warn;
    nu->sp_expire  = rq->sp_expire;
    nu->sp_inact   = rq->sp_inact;

    nu->vs_uuid    = HX_strdup(rq->vs_uuid);
    nu->vs_pvgrp   = HX_strdup(rq->vs_pvgrp);

    HXdeque_push(state->dq_user, nu);
    TOUCH_USER_TAG(1);
    db_flush(state, FLUSH_AUTO);
    return 1;
}

static int vshadow_usermod(struct vxpdb_state *this,
 const struct vxpdb_user *sr_mask, const struct vxpdb_user *mod_mask)
{
#define UP_INT(__field) \
    if(mod_mask->__field != PDB_NO_CHANGE) user->__field = mod_mask->__field;
#define UP_STR(__field) \
    if(mod_mask->__field != NULL) HX_strclone(&user->__field, mod_mask->__field);

    struct shadow_state *state = this->state;
    struct vxpdb_user *user;

    RWLOCK_CHK(state);

    if(sr_mask->pw_name == NULL && sr_mask->pw_uid == PDB_NOUID)
        return -EINVAL;
    user = lookup_user(state->dq_user, sr_mask->pw_name, sr_mask->pw_uid);
    if(user == NULL)
        return -ENOENT;

    UP_STR(pw_name);
    UP_INT(pw_uid);
    UP_INT(pw_gid);
    UP_STR(pw_real);
    UP_STR(pw_home);
    UP_STR(pw_shell);

    UP_STR(sp_passwd);
    UP_INT(sp_lastchg);
    UP_INT(sp_min);
    UP_INT(sp_max);
    UP_INT(sp_warn);
    UP_INT(sp_expire);
    UP_INT(sp_inact);

    UP_STR(vs_uuid);
    UP_STR(vs_pvgrp);
    UP_INT(vs_defer);

    TOUCH_USER_TAG(1);
    db_flush(state, FLUSH_AUTO);
    return 1;
#undef UP_INT
#undef UP_STR
}

static int vshadow_userdel(struct vxpdb_state *this,
 const struct vxpdb_user *sr_mask)
{
    struct shadow_state *state = this->state;
    struct vxpdb_user *user;

    RWLOCK_CHK(state);

    if(sr_mask->pw_name == NULL && sr_mask->pw_uid == PDB_NOUID)
        return -EINVAL;
    user = lookup_user(state->dq_user, sr_mask->pw_name, sr_mask->pw_uid);
    if(user == NULL)
        return -ENOENT;

    HXdeque_del(HXdeque_find(state->dq_user, user));
    free_single_user(user);
    TOUCH_USER_TAG(1);
    db_flush(state, FLUSH_AUTO);
    return 1;
}

static int vshadow_userinfo(struct vxpdb_state *this,
 const struct vxpdb_user *sr_mask, struct vxpdb_user *dest, size_t size)
{
    struct shadow_state *state = this->state;
    const struct HXdeque_node *travp = state->dq_user->first;
    struct vxpdb_user temp_mask;
    int found = 0;

    if(sr_mask == dest) {
        memcpy(&temp_mask, sr_mask, sizeof(struct vxpdb_user));
        sr_mask = &temp_mask;
    }

    /*  Original condition:
            while(... && (dest == NULL || (dest != NULL && size > 0)))

        Table:
        r      s    action
        NULL   ==0  proceed
        NULL   >0   proceed
        !NULL  ==0  stop
        !NULL  >0   proceed

        New formula:
            while(... && !(r != NULL && s == 0))
        Transformed:
            while(... && (r == NULL || s > 0))
    */
    while(travp != NULL && (dest == NULL || size > 0)) {
        const struct vxpdb_user *src = travp->ptr;

        if(!vxpdb_user_match(src, sr_mask)) {
            travp = travp->Next;
            continue;
        }

        if(dest != NULL) {
            vxpdb_user_copy(dest, src);
            ++dest;
            ++found;
            --size;
        } else {
            if(size == 0)
                return 1;
            ++found;
        }
        travp = travp->Next;
    }

    return found;
}

static void *vshadow_usertrav_init(struct vxpdb_state *this) {
    struct shadow_state *state = this->state;
    struct traverser_state trav;

    trav.wp = skip_nis_users(state->dq_user->first);
    return HX_memdup(&trav, sizeof(trav));
}

static int vshadow_usertrav_walk(struct vxpdb_state *this, void *ptr,
 struct vxpdb_user *dest)
{
    struct traverser_state *trav = ptr;
    if(trav->wp == NULL)
        return 0;
    vxpdb_user_copy(dest, trav->wp->ptr);
    trav->wp = skip_nis_users(trav->wp->Next);
    return 1;
}

static void vshadow_usertrav_free(struct vxpdb_state *this, void *ptr) {
    free(ptr);
    return;
}

//-----------------------------------------------------------------------------
static int vshadow_groupadd(struct vxpdb_state *this,
 const struct vxpdb_group *rq)
{
    struct shadow_state *state = this->state;
    struct vxpdb_group *ng;

    if(rq->gr_name == NULL)
        return -EINVAL;
    if(lookup_group(state->dq_group, rq->gr_name, PDB_NOGID) == NULL)
        return -ENOENT;

    RWLOCK_CHK(state);
    if((ng = calloc(1, sizeof(struct vxpdb_group))) == NULL)
        return -errno;

    ng->gr_name = HX_strdup(rq->gr_name);
    ng->gr_gid  = automatic_gid(state, rq->gr_gid);
    HXdeque_push(state->dq_group, ng);
    TOUCH_GROUP_TAG(1);
    db_flush(state, FLUSH_AUTO);
    return 1;
}

static int vshadow_groupmod(struct vxpdb_state *this,
 const struct vxpdb_group *sr_mask, const struct vxpdb_group *mod_mask)
{
    struct shadow_state *state = this->state;
    struct vxpdb_group *group;

    RWLOCK_CHK(state);
    group = lookup_group(state->dq_group, sr_mask->gr_name, sr_mask->gr_gid);
    if(group == NULL)
        return -ENOENT;

    if(mod_mask->gr_name != NULL)
        HX_strclone(&group->gr_name, mod_mask->gr_name);
    if(mod_mask->gr_gid != PDB_NOGID)
        group->gr_gid = mod_mask->gr_gid;

    TOUCH_GROUP_TAG(1);
    db_flush(state, FLUSH_AUTO);
    return 1;
}

static int vshadow_groupdel(struct vxpdb_state *this,
 const struct vxpdb_group *sr_mask)
{
    struct shadow_state *state = this->state;
    struct vxpdb_group *grp;

    RWLOCK_CHK(state);
    grp = lookup_group(state->dq_group, sr_mask->gr_name, sr_mask->gr_gid);
    if(grp == NULL)
        return -ENOENT;

    HXdeque_del(HXdeque_find(state->dq_group, grp));
    free_single_group(grp);
    TOUCH_GROUP_TAG(1);
    db_flush(state, FLUSH_AUTO);
    return 1;
}

static int vshadow_groupinfo(struct vxpdb_state *this,
 const struct vxpdb_group *sr_mask, struct vxpdb_group *dest, size_t size)
{
    struct shadow_state *state = this->state;
    const struct HXdeque_node *travp = state->dq_group->first;
    struct vxpdb_group temp_mask;
    int found = 0;

    if(sr_mask == dest) {
        memcpy(&temp_mask, sr_mask, sizeof(struct vxpdb_group));
        sr_mask = &temp_mask;
    }

    while(travp != NULL && (dest == NULL || size > 0)) {
        const struct vxpdb_group *src = travp->ptr;

        if(!vxpdb_group_match(src, sr_mask)) {
            travp = travp->Next;
            continue;
        }

        if(dest != NULL) {
            vxpdb_group_copy(dest, src);
            ++dest;
            ++found;
            --size;
        } else {
            return 1;
        }
        travp = travp->Next;
    }

    return found;
}

static void *vshadow_grouptrav_init(struct vxpdb_state *this) {
    struct shadow_state *state = this->state;
    struct traverser_state trav;

    trav.wp = skip_nis_groups(state->dq_group->first);
    return HX_memdup(&trav, sizeof(trav));
}

static int vshadow_grouptrav_walk(struct vxpdb_state *this, void *ptr,
 struct vxpdb_group *dest)
{
    struct traverser_state *trav = ptr;
    if(trav->wp == NULL)
        return 0;
    vxpdb_group_copy(dest, trav->wp->ptr);
    trav->wp = skip_nis_groups(trav->wp->Next);
    return 1;
}

static void vshadow_grouptrav_free(struct vxpdb_state *this, void *ptr) {
    free(ptr);
    return;
}

//-----------------------------------------------------------------------------
/* The VXDB_SHADOW locking scheme:
- databases are opened in read mode by default, no shared/read locks
- upon a write attempt (shadow_useradd, etc.) reopen in read-write mode:
  steps: close db, reopen with rwlock, reread, perform operation
- on db_flush(), write db to disk and release write lock (no close/reread)
*/

static int db_open(struct shadow_state *state, int flags) {
#define open_fd(n) ((state->n.fd = open(state->n.path, flags)) >= 0)
#define open_fp(n) ((state->n.fp = fdopen(state->n.fd, fpmode)) != NULL)
    const char *fpmode = (flags & PDB_WRLOCK) ? "r+" : "r";
    int ret;

    if(!open_fd(fpasswd) || !open_fd(fgroup))
        goto out;

    /* Normal users do not get access to /etc/shadow - but should not be
    blamed for it. Only error out when someone tries to write to it.
    "root" usually succeeds. */
    if(!open_fd(fshadow) && (errno != EACCES || flags != O_RDONLY))
        goto out;

    /* /etc/vxshadow is an extension, and it is legal to be missing. It is
    also legal to be non-world-readable. */
    if(!open_fd(fvxshadow) && errno != ENOENT &&
     (errno != EACCES || flags != O_RDONLY))
        goto out;

    if(flags == O_RDWR) {
        struct flock lk = {
            .l_type   = F_WRLCK,
            .l_whence = SEEK_SET,
            .l_start  = 0,
            .l_len    = 0,
        };
        
        if(fcntl(state->fpasswd.fd, F_SETLK, &lk) < 0) {
            fprintf(stderr, "db_shadow: passwd file (%s) currently "
                    "write-locked, waiting\n", state->fpasswd.path);
            if(fcntl(state->fpasswd.fd, F_SETLKW, &lk) < 0)
                goto out;
        }
    }

    if(!open_fp(fpasswd) || !open_fp(fgroup) ||
     (state->fshadow.fd > 0 && !open_fp(fshadow)) ||
     (state->fvxshadow.fd > 0 && !open_fp(fvxshadow)))
        goto out;

    state->dq_user  = db_read_passwd(state->fpasswd.fp);
    state->dq_group = db_read_groups(state->fgroup.fp);

    if(state->fshadow.fp != NULL)
        db_read_shadow(state->fshadow.fp, state->dq_user);
    if(state->fvxshadow.fp != NULL)
        db_read_vxshadow(state->fvxshadow.fp, state->dq_user);

    state->flags = flags;
    return 1;

 out:
    ret = errno;
    db_close(state);
    return -ret;
#undef open_fd
#undef open_fp
}

static void db_close(struct shadow_state *state) {
#define close_fpd(n) \
    if(state->n.fp != NULL) { \
        fclose(state->n.fp); \
        state->n.fd = -1; \
    } \
    if(state->n.fd >= 0) close(state->n.fd);

    if(state->fpasswd.is_chg | state->fshadow.is_chg |
     state->fvxshadow.is_chg | state->fgroup.is_chg)
        db_flush(state, 1);

    close_fpd(fvxshadow);
    close_fpd(fgroup);
    close_fpd(fshadow);
    close_fpd(fpasswd);
    return;
#undef close_fpd
}

/*  db_read_passwd
    @fp:        stdio filehandle to read from

    In this function, we read all users into memory. This speeds processing up,
    as seeking and scanning in userinfo() / groupinfo() would take long
    otherwise. We also have it handy in a struct. Doing so also allows us to
    commit all changes to the user database at once (the usual case with no
    PDB_SYNC flag set).
*/
static struct HXdeque *db_read_passwd(FILE *fp) {
    struct HXdeque *dq;
    struct vxpdb_user *u;
    char *ln = NULL;

    if((dq = HXdeque_init()) == NULL)
        return NULL;

    while(HX_getl(&ln, fp) != NULL) {
        char *data[8];

        if(*ln == '#')
            continue;

        HX_chomp(ln);
        memset(data, 0, sizeof(data));
        if(HX_split5(ln, ":", ARRAY_SIZE(data), data) < 4) {
            fprintf(stderr, "vxdb_shadow: bogus line in passwd file\n");
            continue;
        }

        if((u = malloc(sizeof(struct vxpdb_user))) == NULL)
            break;

        vxpdb_user_clean(u);
        u->pw_name  = HX_strdup(data[0]);
        u->pw_uid   = strtol(data[2], NULL, 0);
        u->pw_gid   = strtol(data[3], NULL, 0);
        u->pw_real  = HX_strdup(data[4]);
        u->pw_home  = HX_strdup(data[5]);
        u->pw_shell = HX_strdup(data[6]);

        /* Anything after the seven main fields is "reserved for future use",
        (read: or private use), so we need to keep that intact. */
        u->be_priv = calloc(1, sizeof(void *) * 2);
        ((char **)u->be_priv)[0] = data[7];

        /* In case there is no shadow entry, we need some recovery values.
        This is already done by vxpdb_clean_user(). */

        /* Stuff is pushed in a linked list. The list is unordered, so it has
        the same order as /etc/passwd. This is intentional (and suffices). */
        HXdeque_push(dq, u);
    }

    hmc_free(ln);
    return dq;
}

/*  db_read_shadow
    @fp:        stdio filehandle to read from
    @dq:        state->dq_user

    Updates the user list with bits from the shadow file.
*/
static void db_read_shadow(FILE *fp, struct HXdeque *dq) {
    struct vxpdb_user *u = NULL;
    char *ln = NULL;

    if(dq->itemcount == 0) // No uses in @dq, so no need to read shadow.
        return;

    while(HX_getl(&ln, fp) != NULL) {
        char *data[9];

        if(*ln == '#')
            continue;

        HX_chomp(ln);
        memset(data, 0, sizeof(data));
        HX_split5(ln, ":", ARRAY_SIZE(data), data);
        if(*data[0] == '\0' || (u = lookup_user(dq, data[0], PDB_NOUID)) == NULL)
            continue; // orphaned entry

        u->sp_passwd  = HX_strdup(data[1]);
        u->sp_lastchg = strtol(data[2], NULL, 0);
        u->sp_min     = strtol(data[3], NULL, 0);
        u->sp_max     = strtol(data[4], NULL, 0);
        u->sp_warn    = strtol(data[5], NULL, 0);

        if(data[6] == NULL || *data[6] == '\0')
            u->sp_expire = strtol(data[6], NULL, 0);
        if(data[7] == NULL || *data[7] == '\0')
            u->sp_inact = strtol(data[7], NULL, 0);
        ((char **)u->be_priv)[1] = HX_strdup(data[8]);
    }

    hmc_free(ln);
    return;
}

static void db_read_vxshadow(FILE *fp, struct HXdeque *dq) {
    struct vxpdb_user *u;
    char *ln = NULL;

    if(dq->itemcount == 0)
        return;

    while(HX_getl(&ln, fp) != NULL) {
        char *data[5];

        if(*ln == '#')
            continue;

        HX_chomp(ln);
        memset(data, 0, sizeof(data));
        HX_split5(ln, ":", ARRAY_SIZE(data), data);
        if(*data[0] == '\0' || (u = lookup_user(dq, data[0], PDB_NOUID)) == NULL)
            continue;

        u->vs_uuid  = HX_strdup(data[1]);
        u->vs_pvgrp = HX_strdup(data[2]);
        u->vs_defer = strtol(data[3], NULL, 0);
    }

    hmc_free(ln);
    return;
}

static struct HXdeque *db_read_groups(FILE *fp) {
    struct vxpdb_group *g;
    struct HXdeque *dq;
    char *ln = NULL;

    if((dq = HXdeque_init()) == NULL)
        return NULL;

    while(HX_getl(&ln, fp) != NULL) {
        char *data[4];

        if(*ln == '#')
            continue;

        HX_chomp(ln);
        memset(data, 0, sizeof(data));
        if(HX_split5(ln, ":", ARRAY_SIZE(data), data) < 3) {
            fprintf(stderr, "vxdb_shadow: bogus line in group file\n");
            continue;
        }

        if((g = malloc(sizeof(struct vxpdb_group))) == NULL)
            return NULL;

        g->gr_name = HX_strdup(data[0]);
        g->gr_gid  = strtol(data[2], NULL, 0);
        g->be_priv = calloc(1, sizeof(void *) * 2);
        ((char **)g->be_priv)[0] = HX_strdup(data[3]);
        HXdeque_push(dq, g);
    }

    hmc_free(ln);
    return dq;
}

/*  db_flush
    @state:     Current instance
    @force:     Force flush even if in-memory copy is recent

    Flush the in-memory database to disk if required. The files must be open
    with %O_RDWR when this function is called.
*/
static void db_flush(struct shadow_state *state, int force) {
    if(!force && !(state->flags & PDB_SYNC))
        return;

    if(state->fpasswd.is_chg | state->fshadow.is_chg |
     state->fvxshadow.is_chg | force)
        db_flush_users(state);

    if(state->fgroup.is_chg | force)
        db_flush_groups(state);

    return;
}

static void db_flush_users(struct shadow_state *state) {
    const struct HXdeque_node *travp = state->dq_user->first;
    int has_vx = state->fvxshadow.fp != NULL;

    fseek(state->fpasswd.fp, 0, SEEK_SET);
    fseek(state->fshadow.fp, 0, SEEK_SET);
    if(has_vx) {
        fseek(state->fvxshadow.fp, 0, SEEK_SET);
        fprintf(state->fvxshadow.fp, "# user:uuid:pvgrp:\n");
    }

    while(travp != NULL) {
        const struct vxpdb_user *u = travp->ptr;
        db_write_passwd(state->fpasswd.fp, u);
        db_write_shadow(state->fshadow.fp, u);
        if(has_vx)
            db_write_vxshadow(state->fvxshadow.fp, u);
        travp = travp->Next;
    }

    truncate_here(state->fpasswd.fp);
    truncate_here(state->fshadow.fp);
    if(has_vx)
        truncate_here(state->fvxshadow.fp);

    TOUCH_USER_TAG(0);
    return;
}

static void db_flush_groups(struct shadow_state *state) {
    const struct HXdeque_node *travp = state->dq_group->first;
    FILE *fp = state->fgroup.fp;

    fseek(state->fgroup.fp, 0, SEEK_SET);

    while(travp != NULL) {
        const struct vxpdb_group *g = travp->ptr;
        const char **pr = g->be_priv;

        fprintf(fp, "%s:x:%ld:", g->gr_name, g->gr_gid);
        if(pr != NULL && pr[0] != NULL)
            fprintf(fp, "%s", pr[0]);
        fprintf(fp, "\n");
        travp = travp->Next;
    }

    truncate_here(state->fgroup.fp);
    TOUCH_GROUP_TAG(0);
    return;
}

static void db_write_passwd(FILE *fp, const struct vxpdb_user *u) {
    const char **priv = u->be_priv;

    fprintf(fp, "%s:x:%ld:%ld:%s:%s:%s",
      u->pw_name, u->pw_uid, u->pw_gid, u->pw_real, u->pw_home, u->pw_shell);
    if(priv != NULL && priv[0] != NULL)
        fprintf(fp, "%s", priv[0]);
    fprintf(fp, "\n");
    return;
}

static void db_write_shadow(FILE *fp, const struct vxpdb_user *u) {
    const char *password = (u->sp_passwd != NULL) ? u->sp_passwd : "!";
    const char **priv = u->be_priv;

    fprintf(fp, "%s:%s:%ld:%ld:%ld:%ld:", u->pw_name, password,
            u->sp_lastchg, u->sp_min, u->sp_max, u->sp_warn);

    if(u->sp_expire != PDB_NO_EXPIRE)   fprintf(fp, "%ld", u->sp_expire);
    fprintf(fp, ":");
    if(u->sp_inact != PDB_NO_INACTIVE)  fprintf(fp, "%ld", u->sp_inact);
    fprintf(fp, ":");
    if(priv != NULL && priv[1] != NULL) fprintf(fp, "%s", priv[1]);
    fprintf(fp, "\n");
    return;
}

static inline void db_write_vxshadow(FILE *fp, const struct vxpdb_user *u) {
    if(u->vs_uuid == NULL && u->vs_pvgrp == NULL && u->vs_defer == 0)
        return;
    fprintf(fp, "%s:%s:%s:%ld:\n", u->pw_name, azstr(u->vs_uuid),
            azstr(u->vs_pvgrp), u->vs_defer);
    return;
}

//-----------------------------------------------------------------------------
static long automatic_uid(struct shadow_state *state, long wanted) {
    const struct HXdeque_node *travp = state->dq_user->first;
    long accept, high = PDB_NOUID, min, max;

    if(wanted == PDB_AUTOUID_SYS) {
        min = accept = 1;
        max = state->uid_min - 1;
    } else if(wanted == PDB_AUTOUID) {
        min = accept = state->uid_min;
        max = state->uid_max;
    } else {
        return wanted;
    }

    // Find the highest ID
    while(travp != NULL) {
        const struct vxpdb_user *user = travp->ptr;
        long uid = user->pw_uid;
        if(uid >= min && uid <= max && uid > high)
            high = uid;
        travp = travp->Next;
    }

    // If the successor id is free, take it.
    if(high >= 0 && high < max)
        return high + 1;

    // No? Then just scan for the next
    while(accept <= max) {
        int used = 0;
        travp = state->dq_user->first;
        while(travp != NULL) {
            const struct vxpdb_user *user = travp->ptr;
            if(user->pw_uid == accept) {
                ++used;
                break;
            }
            travp = travp->Next;
        }
        if(!used) return accept;
        ++accept;
    }
    return -ENOENT;
}

static long automatic_gid(struct shadow_state *state, long wanted) {
    const struct HXdeque_node *travp = state->dq_group->first;
    long accept, high = PDB_NOGID, min, max;

    if(wanted == PDB_AUTOGID_SYS) {
        min = accept = 1;
        max = state->gid_min - 1;
    } else if(wanted == PDB_AUTOGID) {
        min = accept = state->gid_min;
        max = state->gid_max;
    } else {
        return wanted;
    }

    // Find the highest GID
    while(travp != NULL) {
        const struct vxpdb_group *group = travp->ptr;
        long gid = group->gr_gid;
        if(gid >= min && gid <= max && gid > high)
            high = gid;
        travp = travp->Next;
    }

    /* If the highest id is below the max, there must be something free. Take
    the successor id. */
    if(high >= 0 && high < max)
        return high + 1;

    // No? Then just scan for the next
    while(accept <= max) {
        int used = 0;
        travp = state->dq_group->first;
        while(travp != NULL) {
            const struct vxpdb_group *group = travp->ptr;
            if(group->gr_gid == accept) {
                ++used;
                break;
            }
            travp = travp->Next;
        }
        if(!used) return accept;
        ++accept;
    }
    return -ENOENT;
}

static void free_data(struct shadow_state *state) {
    struct HXdeque_node *travp;

    if(state->dq_user != NULL) {
        travp = state->dq_user->first;
        while(travp != NULL) {
            free_single_user(travp->ptr);
            travp = travp->Next;
        }
        HXdeque_free(state->dq_user);
        state->dq_user = NULL;
    }

    if(state->dq_group != NULL) {
        travp = state->dq_group->first;
        while(travp != NULL) {
            free_single_group(travp->ptr);
            travp = travp->Next;
        }
        HXdeque_free(state->dq_group);
        state->dq_group = NULL;
    }
    return;
}

static inline void free_single_user(struct vxpdb_user *u) {
    char **p = u->be_priv;

    free(u->pw_name);
    free(u->pw_real);
    free(u->pw_home);
    free(u->pw_shell);
    free(u->sp_passwd);

    if(p != NULL) {
        free(p[0]);
        free(p[1]);
        free(p);
    }

    free(u);
    return;
}

static inline void free_single_group(struct vxpdb_group *g) {
    char **p = g->be_priv;
    free(g->gr_name);
    if(p != NULL) {
        free(p[0]);
        free(p[1]);
        free(p);
    }
    free(g);
    return;
}

/*  lookup_group
    @dq:        list to search
    @gname:     group name to search on, or %NULL if don't-care
    @gid:       group id to search on, or %PDB_NOGID if don't-care

    Searches the group and returns a pointer to vxdb_shadow's internal data
    structure if found, or %NULL if not found. If @gname is %NULL and @gid is
    %PDB_NOGID, which would usually match any user, %NULL is returned too.
*/
static inline struct vxpdb_group *lookup_group(struct HXdeque *dq,
 const char *gname, long gid)
{
    const struct HXdeque_node *travp = dq->first;

    if(gname == NULL && gid == PDB_NOGID)
        return NULL;

    while(travp != NULL) {
        struct vxpdb_group *g = travp->ptr;
        if((gname == NULL || strcmp(g->gr_name, gname) == 0) &&
         (gid == PDB_NOGID || g->gr_gid == gid))
            return g;
        travp = travp->Next;
    }

    return NULL;
}

/*  lookup_user
    @dq:        list to search on
    @lname:     user login name to match on, or %NULL if don't-care
    @uid:       user uid to match o, or %PDB_NOUID if don't-care

    Searches the user and returns a pointer to vxdb_shadow's internal data
    structure if found, or %NULL if not found. If @lname is %NULL and @uid is
    %PDB_NOUID, which would usually match any user, %NULL is returned too.
*/
static inline struct vxpdb_user *lookup_user(struct HXdeque *dq,
 const char *lname, long uid)
{
    const struct HXdeque_node *travp = dq->first;

    if(lname == NULL && uid == PDB_NOUID)
        return NULL;

    while(travp != NULL) {
        struct vxpdb_user *u = travp->ptr;
        if((lname == NULL || strcmp(u->pw_name, lname) == 0) &&
         (uid == PDB_NOUID || uid == u->pw_uid))
            return u;
        travp = travp->Next;
    }

    return NULL;
}

static void read_config(struct shadow_state *state, unsigned int action) {
    struct HXoption autouid_table[] = {
        {.ln = "UID_MIN", .type = HXTYPE_LONG, .ptr = &state->uid_min},
        {.ln = "UID_MAX", .type = HXTYPE_LONG, .ptr = &state->uid_max},
        {.ln = "GID_MIN", .type = HXTYPE_LONG, .ptr = &state->gid_min},
        {.ln = "GID_MAX", .type = HXTYPE_LONG, .ptr = &state->gid_max},
        HXOPT_TABLEEND,
    };
    struct HXoption options_table[] = {
        {.ln = "PASSWD_DB",   .type = HXTYPE_STRING, .ptr = &state->fpasswd.path},
        {.ln = "SHADOW_DB",   .type = HXTYPE_STRING, .ptr = &state->fshadow.path},
        {.ln = "VXPASSWD_DB", .type = HXTYPE_STRING, .ptr = &state->fvxpasswd.path},
        {.ln = "VXSHADOW_DB", .type = HXTYPE_STRING, .ptr = &state->fvxshadow.path},
        {.ln = "GROUP_DB",    .type = HXTYPE_STRING, .ptr = &state->fgroup.path},
        HXOPT_TABLEEND,
    };

    if(action == CONFIG_READ) {
        state->uid_min = state->gid_min = 1000;
        state->uid_max = state->gid_max = 60000;
        HX_shconfig(CONFIG_ETC_VITALNIX "/autouid.conf", autouid_table);
        HX_shconfig(CONFIG_ETC_VITALNIX "/db_shadow.conf", options_table);
    } else if(action == CONFIG_FREE) {
        HX_shconfig_free(options_table);
    }

    return;
}

static struct HXdeque_node *skip_nis_users(struct HXdeque_node *node) {
    const struct vxpdb_user *user;
    while(node != NULL) {
        user = node->ptr;
        if(*user->pw_name != '+' && *user->pw_name != '-')
            return node;
        node = node->Next;
    }
    return node;
}

static struct HXdeque_node *skip_nis_groups(struct HXdeque_node *node) {
    const struct vxpdb_group *group;
    while(node != NULL) {
        group = node->ptr;
        if(*group->gr_name != '+' && *group->gr_name != '-')
            return node;
        node = node->Next;
    }
    return node;
}

static void truncate_here(FILE *fp) {
    long p = ftell(fp);
    fseek(fp, 0, SEEK_SET);
    fflush(fp);
    ftruncate(fileno(fp), p);
    fflush(fp);
    fseek(fp, p, SEEK_SET);
    return;
}

//=============================================================================
