/*=============================================================================
Vitalnix User Management Suite
libvxpdb/aux.c - auxiliary functions
  Copyright © Jan Engelhardt <jengelh [at] gmx de>, 2005 - 2006
  -- License restrictions apply (LGPL v2.1)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA

  -- For details, see the file named "LICENSE.LGPL2"
=============================================================================*/
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX.h>
#include "libvxpdb/libvxpdb.h"
#include "libvxutil/libvxutil.h"

#define NDSTRLEN(s) (((s) != NULL) ? strlen(s) + 1 : 0)

//-----------------------------------------------------------------------------
void *vxpdb_user_alloc(struct vxpdb_user *user, size_t size) {
    if(size == 0)
        size = 256;
    return hmc_memasg(&user->_storage, NULL, size);
}

void vxpdb_user_clean(struct vxpdb_user *u) {
    memset(u, 0, sizeof(struct vxpdb_user));
    u->pw_uid    = PDB_NOUID;
    u->pw_gid    = PDB_NOGID;
    u->sp_min    = PDB_DFL_KEEPMIN;
    u->sp_max    = PDB_DFL_KEEPMAX;
    u->sp_warn   = PDB_DFL_WARNAGE;
    u->sp_expire = PDB_NO_EXPIRE;
    u->sp_inact  = PDB_NO_INACTIVE;
    return;
}

void vxpdb_user_copy(struct vxpdb_user *dest, const struct vxpdb_user *src) {
    size_t rem;
    hmc_t *wp;

    rem = NDSTRLEN(src->pw_name) + NDSTRLEN(src->pw_real) +
          NDSTRLEN(src->pw_home) + NDSTRLEN(src->pw_shell) +
          NDSTRLEN(src->pw_igrp) /* + NDSTRLEN(src->pw_sgrp) */ +
          NDSTRLEN(src->sp_passwd) + NDSTRLEN(src->vs_uuid) +
          NDSTRLEN(src->vs_pvgrp);
    wp  = vxpdb_user_alloc(dest, rem);

    dest->pw_name    = vxutil_zcat(&wp, src->pw_name,   &rem);
    dest->pw_uid     = src->pw_uid;
    dest->pw_gid     = src->pw_gid;
    dest->pw_igrp    = vxutil_zcat(&wp, src->pw_igrp,   &rem);
    dest->pw_real    = vxutil_zcat(&wp, src->pw_real,   &rem);
    dest->pw_home    = vxutil_zcat(&wp, src->pw_home,   &rem);
    dest->pw_shell   = vxutil_zcat(&wp, src->pw_shell,  &rem);
    dest->sp_passwd  = vxutil_zcat(&wp, src->sp_passwd, &rem);
    dest->sp_lastchg = src->sp_lastchg;
    dest->sp_min     = src->sp_min;
    dest->sp_max     = src->sp_max;
    dest->sp_warn    = src->sp_warn;
    dest->sp_expire  = src->sp_expire;
    dest->sp_inact   = src->sp_inact;
    dest->vs_defer   = src->vs_defer;
    dest->vs_uuid    = vxutil_zcat(&wp, src->vs_uuid,  &rem);
    dest->vs_pvgrp   = vxutil_zcat(&wp, src->vs_pvgrp, &rem);
    dest->be_priv    = NULL;
    return;
}

struct vxpdb_user *vxpdb_user_dup(const struct vxpdb_user *src) {
    struct vxpdb_user *dest;
    if((dest = malloc(sizeof(struct vxpdb_user))) == NULL)
        return NULL;
    dest->_storage = NULL;
    vxpdb_user_copy(dest, src);
    return dest;
}

void vxpdb_user_free(struct vxpdb_user *user, int heap) {
    hmc_free(user->_storage);
    if(heap)
        free(user);
    else
        user->_storage = NULL;
    return;
}

int vxpdb_user_match(const struct vxpdb_user *user,
 const struct vxpdb_user *mask)
{
    return
      (mask->pw_name == NULL || strcmp(user->pw_name, mask->pw_name) == 0) &&
      (mask->pw_uid == PDB_NOUID || user->pw_uid == mask->pw_uid) &&
      (mask->pw_gid == PDB_NOGID || user->pw_gid == mask->pw_gid) &&
      (mask->pw_real == NULL || strcmp(user->pw_real, mask->pw_real) == 0) &&
      (mask->pw_home == NULL || strcmp(user->pw_home, mask->pw_home) == 0) &&
      (mask->pw_shell == NULL || strcmp(user->pw_shell, mask->pw_shell) == 0);
    // lacks comparing mask->{sp,md}_*
}

void vxpdb_user_nomodify(struct vxpdb_user *u) {
    memset(u, 0, sizeof(struct vxpdb_user));
    u->pw_uid     = PDB_NO_CHANGE;
    u->pw_gid     = PDB_NO_CHANGE;
    u->sp_lastchg = PDB_NO_CHANGE;
    u->sp_min     = PDB_NO_CHANGE;
    u->sp_max     = PDB_NO_CHANGE;
    u->sp_warn    = PDB_NO_CHANGE;
    u->sp_expire  = PDB_NO_CHANGE;
    u->sp_inact   = PDB_NO_CHANGE;
    return;
}

void *vxpdb_group_alloc(struct vxpdb_group *group, size_t size) {
    if(size == 0)
        size = 256;
    return hmc_memasg(&group->_storage, NULL, size);
}

void vxpdb_group_clean(struct vxpdb_group *group) {
    group->gr_name = NULL;
    group->gr_gid  = PDB_NOGID;
    group->be_priv = NULL;
    return;
}

void vxpdb_group_copy(struct vxpdb_group *dest, const struct vxpdb_group *src) {
    size_t rem;
    hmc_t *wp;

    rem = NDSTRLEN(src->gr_name);
    wp  = vxpdb_group_alloc(dest, rem);

    dest->gr_name = vxutil_zcat(&wp, src->gr_name, &rem);
    dest->gr_gid  = src->gr_gid;
    dest->be_priv = NULL;
    return;
}

struct vxpdb_group *vxpdb_group_dup(const struct vxpdb_group *src) {
    struct vxpdb_group *dest;
    if((dest = malloc(sizeof(struct vxpdb_group))) == NULL)
        return NULL;
    dest->_storage = NULL;
    vxpdb_group_copy(dest, src);
    return dest;
}

void vxpdb_group_free(struct vxpdb_group *group, int heap) {
    hmc_free(group->_storage);
    if(heap)
        free(group);
    else
        group->_storage = NULL;
    return;
}

int vxpdb_group_match(const struct vxpdb_group *group,
 const struct vxpdb_group *mask)
{
    return
      (mask->gr_name == NULL || strcmp(group->gr_name, mask->gr_name) == 0) &&
      (mask->gr_gid == PDB_NOGID || group->gr_gid == mask->gr_gid);
}

void vxpdb_group_nomodify(struct vxpdb_group *group) {
    group->gr_name = NULL;
    group->gr_gid  = PDB_NO_CHANGE;
    group->be_priv = NULL;
    return;
}

//=============================================================================
