/*=============================================================================
Vitalnix User Management Suite
steelmill/wd_sync.cpp
  Copyright © Jan Engelhardt <jengelh [at] gmx de>, 2005 - 2006
  -- License restrictions apply (LGPL v2.1)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA

  -- For details, see the file named "LICENSE.LGPL2"
=============================================================================*/
#include <sys/time.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#    include <wx/wx.h>
#endif
#include <wx/spinctrl.h>
#include <wx/statline.h>
#include "libvxeds/libvxeds.h"
#include "libvxeds/vtable.h"
#include "libvxmdsync/libvxmdsync.h"
#include "libvxpdb/libvxpdb.h"
#include "libvxpdb/xafunc.h"
#include "libvxpdb/xwfunc.h"
#include "libvxutil/defines.h"
#include "libvxutil/libvxutil.h"
#include "steelmill/common.hpp"
#include "steelmill/wd_sync.hpp"

enum {
    // StaticText for countdowns
    ID_CD_BASE          = 0x10,
    ID_CD_COMPARE       = ID_CD_BASE | MDREP_COMPARE,
    ID_CD_ADD           = ID_CD_BASE | MDREP_ADD,
    ID_CD_UPDATE        = ID_CD_BASE | MDREP_UPDATE,
    ID_CD_DSTART        = ID_CD_BASE | MDREP_DSTART,
    ID_CD_DWAIT         = ID_CD_BASE | MDREP_DWAIT,
    ID_CD_DSTOP         = ID_CD_BASE | MDREP_DSTOP,
    ID_CD_DELETE        = ID_CD_BASE | MDREP_DELETE,
    // Gauges
    ID_GG_BASE          = 0x20,
    ID_GG_ADD           = ID_GG_BASE | MDREP_ADD,
    ID_GG_UPDATE        = ID_GG_BASE | MDREP_UPDATE,
    ID_GG_DSTART        = ID_GG_BASE | MDREP_DSTART,
    ID_GG_DSTOP         = ID_GG_BASE | MDREP_DSTOP,
    ID_GG_DELETE        = ID_GG_BASE | MDREP_DELETE,
    ID_GG_COMPARE       = ID_GG_BASE | MDREP_COMPARE,
    ID_GG_FIXUP         = ID_GG_BASE | MDREP_FIXUP,
    // StaticText for %
    ID_PT_BASE          = 0x40,
    ID_PT_ADD           = ID_PT_BASE | MDREP_ADD,
    ID_PT_UPDATE        = ID_PT_BASE | MDREP_UPDATE,
    ID_PT_DSTART        = ID_PT_BASE | MDREP_DSTART,
    ID_PT_DSTOP         = ID_PT_BASE | MDREP_DSTOP,
    ID_PT_DELETE        = ID_PT_BASE | MDREP_DELETE,
    ID_PT_COMPARE       = ID_PT_BASE | MDREP_COMPARE,
    ID_PT_FIXUP         = ID_PT_BASE | MDREP_FIXUP,
    // "Details" buttons
    ID_DT_BASE          = 0x80,
    ID_DT_COMPARE       = ID_DT_BASE | MDREP_COMPARE,
    ID_DT_ADD           = ID_DT_BASE | MDREP_ADD,
    ID_DT_UPDATE        = ID_DT_BASE | MDREP_UPDATE,
    ID_DT_DSTART        = ID_DT_BASE | MDREP_DSTART,
    ID_DT_DWAIT         = ID_DT_BASE | MDREP_DWAIT,
    ID_DT_DSTOP         = ID_DT_BASE | MDREP_DSTOP,
    ID_DT_DELETE        = ID_DT_BASE | MDREP_DELETE,
};

struct pv_sync {
    struct vxpdb_state *module_handle;
    struct vxpdb_group group;
    struct mdsync_workspace *mdsw;
    bool prompt;
    int open_status;
    long num_eds_users;
    wxWindow *window;
    struct timeval last_print;
};

class WD_SyncCompare : public wxDialog {
  public: // functions
    WD_SyncCompare(wxWindow *, struct pv_sync *);
};

class WD_SyncProg : public wxDialog {
  public: // functions
    WD_SyncProg(wxWindow *, struct pv_sync *);

  private: // functions
    void Continue(wxCommandEvent &);
    void Details_Add(wxCommandEvent &);
    void Details_Update(wxCommandEvent &);
    void Details_Defer_Start(wxCommandEvent &);
    void Details_Defer_Wait(wxCommandEvent &);
    void Details_Defer_Stop(wxCommandEvent &);
    void Details_Delete(wxCommandEvent &);

  private: // variables
    struct pv_sync *priv;
    DECLARE_EVENT_TABLE();
};

// Functions
static void add_edsformats(wxChoice *);
static void add_groups(wxComboBox *);
static void cb_report(unsigned int, const struct mdsync_workspace *,
    unsigned long, unsigned long);
template<class Object> static inline Object find_window(long, wxWindow *);
static void listbox_fill_bt(wxListBox *, const void *);
static void release_priv(struct pv_sync *);
static int time_limit(struct timeval *, const struct timeval *);
static void tv_delta(const struct timeval *, const struct timeval *,
    struct timeval *);
static void update_max(wxWindow *, long);
static wxString wxString_int(unsigned long);

//-----------------------------------------------------------------------------
BEGIN_EVENT_TABLE(WD_SyncParam, wxDialog)
    EVT_BUTTON(wxID_FORWARD, WD_SyncParam::Compare)
END_EVENT_TABLE()

WD_SyncParam::WD_SyncParam(wxWindow *parent) :
    wxDialog(parent, wxID_ANY, PROD_NAME " > Synchronize",
     wxDPOS, wxDSIZE, wxCFF)
{
    wxBoxSizer *vp = new wxBoxSizer(wxVERTICAL);

    wxStaticBox *sb_src      = new wxStaticBox(this, wxID_ANY, "Source");
    wxStaticBoxSizer *sp_src = new wxStaticBoxSizer(sb_src, wxVERTICAL);
    wxFlexGridSizer *sf_src  = new wxFlexGridSizer(2);
    sf_src->AddGrowableCol(1);
    sf_src->Add(new wxStaticText(this, wxID_ANY, "Resource Identifier:"), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    sf_src->Add(ct_source = new GW_FTC(this, "../../Vitalnix-PRV/daten3.sdf", 3), 0, wxACV | wxGROW);
    sf_src->Add(new wxStaticText(this, wxID_ANY, "Type:"), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    sf_src->Add(ct_srctype = new wxChoice(this, wxID_ANY, wxDPOS, wxDSIZE, 0, NULL), 0, wxACV | wxALL, 3);
    sp_src->Add(sf_src, 1, wxGROW);

    wxStaticBox *sb_dst      = new wxStaticBox(this, wxID_ANY, "Destination");
    wxStaticBoxSizer *sp_dst = new wxStaticBoxSizer(sb_dst, wxVERTICAL);
    wxFlexGridSizer *sf_dst  = new wxFlexGridSizer(2);
    sf_dst->AddGrowableCol(1);
    sf_dst->Add(new wxStaticText(this, wxID_ANY, "Database"), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    sf_dst->Add(ct_module = new wxComboBox(this, wxID_ANY, "*"), 0, wxACV | wxALL, 3);
    sf_dst->Add(new wxStaticText(this, wxID_ANY, "Group or GID"), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    sf_dst->Add(ct_group = new wxComboBox(this, wxID_ANY, wxEmptyString), 0, wxACV | wxALL, 3);
    sp_dst->Add(sf_dst, 1, wxGROW);

    wxStaticBox *sb_out      = new wxStaticBox(this, wxID_ANY, "Output");
    wxStaticBoxSizer *sp_out = new wxStaticBoxSizer(sb_out, wxVERTICAL);
    wxFlexGridSizer *sf_out  = new wxFlexGridSizer(2);
    sf_out->AddGrowableCol(1);
    sf_out->Add(new wxStaticText(this, wxID_ANY, "Logfile:"), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    sf_out->Add(ct_output = new GW_FTC(this, "../../Vitalnix-PRV/daten3.log", 3), 0, wxACV | wxGROW);
    sp_out->Add(sf_out, 1, wxGROW);

    sp_out->Add(ct_prompt = new wxCheckBox(this, wxID_ANY, "Prompt before operations"), 0, wxALIGN_LEFT | wxALL, 3);

    add_edsformats(ct_srctype);
    add_groups(ct_group);
    ct_srctype->SetSelection(0);
    ct_group->SetValue("extern");

    vp->Add(sp_src, 0, wxGROW | wxALL, 5);
    vp->Add(sp_dst, 0, wxGROW | wxALL, 5);
    vp->Add(sp_out, 0, wxGROW | wxALL, 5);
    vp->Add(-1, 30, 1);
    vp->Add(smc_navgen(this, "c-n"), 0, wxACV | wxGROW);

    SetSizer(vp);
    vp->SetSizeHints(this);
    smc_size_minimum(this, 400, 300);
    smc_size_aspect(this);
    Center();

    priv = NULL;
    return;
}

void WD_SyncParam::Compare(wxCommandEvent &event) {
    wxString group_name, input_file, output_file, str;
    const char *format;
    int ret;

    try {
        input_file = ct_source->GetValue();
        if(strcmp(input_file, "") == 0) {
            str = "Data Source may not be empty!";
            throw;
        }
        output_file = ct_output->GetValue();
        if(strcmp(output_file, "") == 0) {
            str = "Output logfile may not be empty!";
            throw;
        }
        if(strcmp(ct_module->GetValue(), "") == 0) {
            str = "Database selector may not be empty!";
            throw;
        }
        group_name = ct_group->GetValue();
        if(strcmp(group_name, "") == 0) {
            str = "Group may not be empty!";
            throw;
        }

        priv = new struct pv_sync;
        memset(priv, 0, sizeof(*priv));

        priv->module_handle = vxpdb_load(ct_module->GetValue(), NULL);
        if(priv->module_handle == NULL) {
            str.Printf("Could not load PDB backend module: %s\n", strerror(errno));
            throw;
        }

        if((ret = vxpdb_open(priv->module_handle, PDB_WRLOCK)) <= 0) {
            str.Printf("Could not open PDB module: %s\n", strerror(-ret));
            throw;
        }

        priv->open_status = 1;
        if((priv->mdsw = mdsync_init()) == NULL) {
            str.Printf("mdsync_init(): %s\n", strerror(errno));
            throw;
        }

        priv->mdsw->database     = priv->module_handle;
        priv->mdsw->report       = cb_report;
        priv->mdsw->user_private = priv;
        priv->window             = this;

        if((ret = mdsync_prepare_group(priv->mdsw, group_name)) < 0) {
            str.Printf("Could not look up group \"%s\": %s",
                static_cast<const char *>(group_name), strerror(-ret));
            throw;
        } else if(ret == 0) {
            str.Printf("Group \"%s\" does not exist",
                static_cast<const char *>(group_name));
            throw;
        }

        format = static_cast<const char *>(ct_srctype->GetClientData(ct_srctype->GetSelection()));
        if(format == NULL && (format = vxeds_derivefromname(input_file)) == NULL) {
            str = "Could not determine file type of input data source.";
            throw;
        }

        if((ret = mdsync_read_file(priv->mdsw, input_file, format)) <= 0) {
            str.Printf("Error while reading Data Source: %s\n", strerror(-ret));
            throw;
        }

        if((ret = mdsync_open_log(priv->mdsw, output_file)) <= 0) {
            str.Printf("Error trying to open logfile: %s\n", strerror(-ret));
            throw;
        }

        priv->prompt = ct_prompt->GetValue();

        WD_SyncCompare cmp_dlg(this, priv);
        cmp_dlg.Show();
        wxTheApp->Yield();
        mdsync_compare(priv->mdsw);
        update_max(&cmp_dlg, MDREP_COMPARE);
        mdsync_fixup(priv->mdsw);
        update_max(&cmp_dlg, MDREP_FIXUP);
        cmp_dlg.Hide();

        Hide();
        WD_SyncProg(this, priv).ShowModal();
        Close(true);
    } catch(...) {
        GD_Message(this, "Error", str, "-o").ShowModal();
    }
    release_priv(priv);
    return;
}

//-----------------------------------------------------------------------------
WD_SyncCompare::WD_SyncCompare(wxWindow *parent, struct pv_sync *priv) :
    wxDialog(parent, wxID_ANY, PROD_NAME " > Synchronize > Compare stage",
     wxDPOS, wxDSIZE, wxCFF)
{
    wxFlexGridSizer *st = new wxFlexGridSizer(3);
    wxBoxSizer *vp      = new wxBoxSizer(wxVERTICAL);

    st->AddGrowableCol(1);

    st->Add(new wxStaticText(this, wxID_ANY, "Compare"), 0, wxALIGN_LEFT | wxACV | wxALL, 3);
    st->Add(new wxGauge(this, ID_GG_COMPARE, 1, wxDPOS, wxDSIZE, wxGA_SMOOTH), 0, wxGROW | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, ID_PT_COMPARE, "0%", wxDPOS, wxSize(6 * GetCharWidth(), -1), wxALIGN_RIGHT), 0, wxALIGN_RIGHT | wxACV | wxALL, 0);

    st->Add(new wxStaticText(this, wxID_ANY, "Fixup"), 0, wxALIGN_LEFT | wxACV | wxALL, 3);
    st->Add(new wxGauge(this, ID_GG_FIXUP, 1, wxDPOS, wxDSIZE, wxGA_SMOOTH), 0, wxGROW | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, ID_PT_FIXUP, "0%"), 0, wxALIGN_RIGHT | wxACV | wxALL, 0);

    vp->Add(new wxStaticText(this, wxID_ANY, "Comparing EDS to PDB... Please wait."), 0, wxALIGN_LEFT | wxALL, 10);
    vp->Add(st, 0, wxGROW | wxLEFT | wxRIGHT | wxBOTTOM, 10);
    vp->Add(new wxStaticLine(this, wxID_ANY), 0, wxGROW | wxTOP | wxBOTTOM, 5);

    SetSizer(vp);
    vp->SetSizeHints(this);
    smc_size_aspect(this, 3);
    Center();

    priv->window = this;
    return;
}

//-----------------------------------------------------------------------------
BEGIN_EVENT_TABLE(WD_SyncProg, wxDialog)
    EVT_BUTTON(wxID_FORWARD, WD_SyncProg::Continue)
    EVT_BUTTON(ID_DT_ADD,    WD_SyncProg::Details_Add)
    EVT_BUTTON(ID_DT_UPDATE, WD_SyncProg::Details_Update)
    EVT_BUTTON(ID_DT_DSTART, WD_SyncProg::Details_Defer_Start)
    EVT_BUTTON(ID_DT_DWAIT,  WD_SyncProg::Details_Defer_Wait)
    EVT_BUTTON(ID_DT_DSTOP,  WD_SyncProg::Details_Defer_Stop)
    EVT_BUTTON(ID_DT_DELETE, WD_SyncProg::Details_Delete)
END_EVENT_TABLE()

WD_SyncProg::WD_SyncProg(wxWindow *parent, struct pv_sync *priv) :
    wxDialog(parent, wxID_ANY, PROD_NAME " > Synchronize > Operation in progress",
     wxDPOS, wxDSIZE, wxCFF)
{
    struct mdsync_workspace *mdsw = priv->mdsw;
    wxFlexGridSizer *st = new wxFlexGridSizer(5);
    wxBoxSizer *vp0     = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *vp1     = new wxBoxSizer(wxVERTICAL);

    st->AddGrowableCol(2);

    vp1->Add(new wxStaticText(this, wxID_ANY,
        "Please wait until the selected operations have been completed.\n"
        "This may take a while depending on your configuration.\n"
        "Process might not occur linearly."),
        0, wxACV | wxALIGN_LEFT | wxALL, 5);

    st->Add(new wxStaticText(this, wxID_ANY, wxString_int(mdsw->num_grp)), 0, wxALIGN_RIGHT | wxALL, 3);
    st->Add(new wxStaticText(this, wxID_ANY, "group member(s) found in PDB"), 0, wxALIGN_LEFT | wxACV | wxALL, 3);
    st->Add(-1, -1);
    st->Add(-1, -1);
    st->Add(-1, -1);

    st->Add(new wxStaticText(this, ID_CD_ADD, wxString_int(mdsw->add_req->itemcount), wxDPOS, wxSize(8 * GetCharWidth(), -1), wxALIGN_RIGHT), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, wxID_ANY, "new users to add"), 0, wxALIGN_LEFT | wxACV | wxALL, 3);
    st->Add(new wxGauge(this, ID_GG_ADD, 1, wxDPOS, wxDSIZE, wxGA_SMOOTH), 0, wxGROW | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, ID_PT_ADD, "0%", wxDPOS, wxSize(6 * GetCharWidth(), -1), wxALIGN_RIGHT), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxButton(this, ID_DT_ADD, "Details"), 0, wxGROW | wxACV | wxALL, 3);

    st->Add(new wxStaticText(this, ID_CD_UPDATE, wxString_int(mdsw->keep_req->itemcount), wxDPOS, wxDSIZE, wxALIGN_RIGHT), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, wxID_ANY, "to keep and update group descriptors"), 0, wxALIGN_LEFT | wxACV | wxALL, 3);
    st->Add(new wxGauge(this, ID_GG_UPDATE, 1, wxDPOS, wxDSIZE, wxGA_SMOOTH), 0, wxGROW | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, ID_PT_UPDATE, "0%"), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxButton(this, ID_DT_UPDATE, "Details"), 0, wxGROW | wxACV | wxALL, 3);

    st->Add(new wxStaticText(this, ID_CD_DSTART, wxString_int(mdsw->defer_start->itemcount)), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, wxID_ANY, "to start deferred deletion timer"), 0, wxALIGN_LEFT | wxACV | wxALL, 3);
    st->Add(new wxGauge(this, ID_GG_DSTART, 1, wxDPOS, wxDSIZE, wxGA_SMOOTH), 0, wxGROW | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, ID_PT_DSTART, "0%"), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxButton(this, ID_DT_DSTART, "Details"), 0, wxGROW | wxACV | wxALL, 3);

    st->Add(new wxStaticText(this, wxID_ANY, wxString_int(mdsw->defer_wait->itemcount)), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, wxID_ANY, "to wait for deletion"), 0, wxALIGN_LEFT | wxACV | wxALL, 3);
    st->Add(-1, -1);
    st->Add(-1, -1);
    st->Add(new wxButton(this, ID_DT_DWAIT, "Details"), 0, wxGROW | wxACV | wxALL, 3);

    st->Add(new wxStaticText(this, ID_CD_DSTOP, wxString_int(mdsw->defer_stop->itemcount)), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, wxID_ANY, "to stop deferred deletion timer"), 0, wxALIGN_LEFT | wxACV | wxALL, 3);
    st->Add(new wxGauge(this, ID_GG_DSTOP, 1, wxDPOS, wxDSIZE, wxGA_SMOOTH), 0, wxGROW | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, ID_PT_DSTOP, "0%"), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxButton(this, ID_DT_DSTOP, "Details"), 0, wxGROW | wxACV | wxALL, 3);

    st->Add(new wxStaticText(this, ID_CD_DELETE, wxString_int(mdsw->delete_now->itemcount)), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, wxID_ANY, "to delete"), 0, wxALIGN_LEFT | wxACV | wxALL, 3);
    st->Add(new wxGauge(this, ID_GG_DELETE, 1, wxDPOS, wxDSIZE, wxGA_SMOOTH), 0, wxGROW | wxACV | wxALL, 3);
    st->Add(new wxStaticText(this, ID_PT_DELETE, "0%"), 0, wxALIGN_RIGHT | wxACV | wxALL, 3);
    st->Add(new wxButton(this, ID_DT_DELETE, "Details"), 0, wxGROW | wxACV | wxALL, 3);

    vp1->Add(st, 0, wxGROW);
    vp0->Add(vp1, 1, wxGROW | wxALL, 20);
    vp0->Add(smc_navgen(this, "c-n"), 0, wxGROW);

    SetSizer(vp0);
    vp0->SetSizeHints(this);
    smc_size_aspect(this, 2);
    Center();

    this->priv   = priv;
    priv->window = this;
    return;
}

void WD_SyncProg::Continue(wxCommandEvent &event) {
    struct mdsync_workspace *mdsw = priv->mdsw;
    wxButton *btn;

    btn = find_window<wxButton *>(wxID_FORWARD, this);
    btn->Disable();

    GD_Message dlg_add(this, PROD_NAME " > Synchronize > Adding users",
        "Continue with adding users?", "-oc-");
    if(mdsw->add_req->itemcount > 0 && (!priv->prompt ||
     dlg_add.ShowModal() == wxID_OK))
            mdsync_add(mdsw);

    update_max(this, MDREP_ADD);

    GD_Message dlg_mod(this, PROD_NAME " > Synchronize > Update timers",
        "Continue with updating the deferred deletion timers?", "-oc");
    if(mdsw->defer_start->itemcount + mdsw->defer_stop->itemcount > 0 &&
     (!priv->prompt || dlg_mod.ShowModal() == wxID_OK))
            mdsync_mod(mdsw);

    update_max(this, MDREP_UPDATE);
    update_max(this, MDREP_DSTART);
    update_max(this, MDREP_DSTOP);

    GD_Message dlg_del(this, PROD_NAME " > Synchronize > Delete users",
        "Continue with deleting old users?", "-oc-");
    if(mdsw->delete_now->itemcount > 0 && (!priv->prompt ||
     dlg_del.ShowModal() == wxID_OK))
            mdsync_del(mdsw);

    update_max(this, MDREP_DELETE);

    btn = find_window<wxButton *>(wxID_CANCEL, this);
    btn->Disable();

    btn = find_window<wxButton *>(wxID_FORWARD, this);
    btn->SetLabel("&Finish");
    btn->SetId(wxID_OK);
    btn->Enable();

    GetSizer()->Layout();
    return;
}

void WD_SyncProg::Details_Update(wxCommandEvent &event) {
    GD_Listbox(this, "", listbox_fill_bt, priv->mdsw->keep_req->root,
        wxLB_SORT).ShowModal();
    return;
}

void WD_SyncProg::Details_Add(wxCommandEvent &event) {
    GD_Listbox(this, "", listbox_fill_bt, priv->mdsw->add_req->root,
        wxLB_SORT).ShowModal();
    return;
}

void WD_SyncProg::Details_Defer_Start(wxCommandEvent &event) {
    GD_Listbox(this, "", priv->mdsw->defer_start, wxLB_SORT).ShowModal();
    return;
}

void WD_SyncProg::Details_Defer_Wait(wxCommandEvent &event) {
    GD_Listbox(this, "", priv->mdsw->defer_wait, wxLB_SORT).ShowModal();
    return;
}

void WD_SyncProg::Details_Defer_Stop(wxCommandEvent &event) {
    GD_Listbox(this, "", priv->mdsw->defer_stop, wxLB_SORT).ShowModal();
    return;
}

void WD_SyncProg::Details_Delete(wxCommandEvent &event) {
    GD_Listbox(this, "", priv->mdsw->delete_now, wxLB_SORT).ShowModal();
    return;
}

//-----------------------------------------------------------------------------
static void add_edsformats(wxChoice *c) {
    const struct edsformat_vtable *vtable;
    void *trav = NULL;

    c->Append("(autodetect)", static_cast<void *>(NULL));
    while((vtable = vxeds_formats_trav(&trav)) != NULL)
        c->Append(vtable->desc, const_cast<struct edsformat_vtable *>(vtable));

    return;
}

static void add_groups(wxComboBox *b) {
    struct vxpdb_group group = {};
    struct vxpdb_state *db;
    void *trav = NULL;

    if((db = vxpdb_load("*", NULL)) == NULL)
        return;
    if(vxpdb_open(db, 0) <= 0)
        goto out_unload;
    if((trav = vxpdb_grouptrav_init(db)) == NULL)
        goto out_close;
    while(vxpdb_grouptrav_walk(db, trav, &group) > 0)
        if(group.gr_gid > 0)
            b->Append(group.gr_name);
    vxpdb_grouptrav_free(db, trav);
    vxpdb_group_free(&group, 0);

 out_close:
    vxpdb_close(db);
 out_unload:
    vxpdb_unload(db);
    return;
}

static void cb_report(unsigned int type, const struct mdsync_workspace *mdsw,
 unsigned long current, unsigned long max)
{
    static const struct timeval report_delta[] = {0, 100000};
    struct pv_sync *priv = static_cast<struct pv_sync *>(mdsw->user_private);
    wxStaticText *cd, *pt;
    wxGauge *gg;

    if(current == 1)
        memset(&priv->last_print, 0, sizeof(priv->last_print));
    if(!time_limit(&priv->last_print, report_delta) && current != max)
        return;

    wxString s;
    cd = static_cast<wxStaticText *>(priv->window->FindWindowById(type | ID_CD_BASE, priv->window));
    pt = static_cast<wxStaticText *>(priv->window->FindWindowById(type | ID_PT_BASE, priv->window));
    gg = static_cast<wxGauge *>(priv->window->FindWindowById(type | ID_GG_BASE, priv->window));

    if(cd != NULL) {
        s.Printf("%ld", max - current);
        cd->SetLabel(s);
    }
    s.Printf("%ld%%", current * 100 / max);
    pt->SetLabel(s);
    if(current == 1)
        gg->SetRange(max);
    gg->SetValue(current);
    priv->window->GetSizer()->Layout();
    wxTheApp->Yield();
    return;
}

template<class Object> static inline Object find_window(long n, wxWindow *w) {
    return static_cast<Object>(w->FindWindowById(n, w));
}

static void listbox_fill_bt(wxListBox *lb, const void *user_ptr) {
    const struct HXbtree_node *node;
    const struct vxeds_entry *e;
    wxString s;

    node = static_cast<const struct HXbtree_node *>(user_ptr);
    if(node == NULL)
        return;

    e    = static_cast<const struct vxeds_entry *>(node->data);
    s.Printf("%s (%s)", e->username, e->full_name);
    lb->Append(s);

    if(node->sd[0] != NULL)
        listbox_fill_bt(lb, node->sd[0]);
    if(node->sd[1] != NULL)
        listbox_fill_bt(lb, node->sd[1]);
    return;
}

static void release_priv(struct pv_sync *priv) {
    if(priv == NULL)
        return;

    vxpdb_group_free(&priv->group, 0);
    if(priv->open_status)
        vxpdb_close(priv->module_handle);
    if(priv->module_handle != NULL)
        vxpdb_unload(priv->module_handle);
    delete priv;
    return;
}

static int time_limit(struct timeval *last, const struct timeval *max) {
    struct timeval now, delta;
    int update;

    gettimeofday(&now, NULL);
    tv_delta(last, &now, &delta);
    update = delta.tv_sec > max->tv_sec ||
             (delta.tv_sec == max->tv_sec && delta.tv_usec > max->tv_usec);

    if(update)
        *last = now;
    return update;
}

static void tv_delta(const struct timeval *past, const struct timeval *now,
 struct timeval *dest)
{
    /* Calculates the time difference between "past" and "now" and stores the
    result in "dest". All parameters in µs. */
    unsigned long sec = now->tv_sec - past->tv_sec;
    long acc = now->tv_usec - past->tv_usec;

    if(acc < 0) {
        // past: 1.5, now: 2.0, sec = 2 - 1 = 1, acc = 0 - 500000 = -500000;
        dest->tv_sec = --sec;
        dest->tv_usec = 1000000 + acc;
    } else {
        dest->tv_sec = sec;
        dest->tv_usec = acc;
    }

    return;
}

static void update_max(wxWindow *w, long id) {
    wxStaticText *pt = find_window<wxStaticText *>(id | ID_PT_BASE, w);
    wxGauge *gg      = find_window<wxGauge *>(id | ID_GG_BASE, w);

    if(pt != NULL)
        pt->SetLabel("100%");
    if(gg != NULL)
        gg->SetValue(gg->GetRange());

    return;
}

static wxString wxString_int(unsigned long n) {
    wxString s;
    s.Printf("%lu", n);
    return s;
}

//=============================================================================
